"""Module with abstract classes for orredis"""
from datetime import datetime, timezone
from typing import Dict, Any

from pydantic import BaseModel


class Model(BaseModel):
    """Base model for all models that can be saved and retrieved from redis using orredis"""

    def with_changes(self, changes: Dict[str, Any]) -> "Model":
        """
        Creates a new instance of Model basing on the properties the current one has
        but with modifications using changes
        """
        return self.__class__(**{**self.dict(), **changes})

    def __eq__(self, other):
        datetime_fields = self.__get_datetime_fields()
        if len(datetime_fields) == 0:
            return super().__eq__(other=other)
        else:
            return self.__eq_with_datetime_fields(datetime_fields=datetime_fields, other=other)

    def __get_datetime_fields(self):
        """Retrieves the fields that have datetime as their field type"""
        datetime_fields = getattr(self.__class__, "__datetime_fields", None)

        if datetime_fields is None:
            datetime_fields = {}
            for (k, v) in self.__fields__.items():
                try:
                    if issubclass(v, datetime):
                        datetime_fields[k] = True
                except:
                    pass

            setattr(self.__class__, "__datetime_fields", datetime_fields)

        return datetime_fields

    def __eq_with_datetime_fields(self, datetime_fields, other):
        """
        compares the self to other to determine equality if datetime_fields are present
        since timezone-aware fields seem to cause wrong behaviour

        :param datetime_fields: the fields that have datetime as their type
        :param other: the model or dict to compare self to
        :return: whether self is equal to other
        """
        current_dict = self.dict()
        other_dict = other.dict() if isinstance(other, BaseModel) else other

        if len(current_dict) != len(other_dict):
            return False

        for key in current_dict:
            try:
                current_item = current_dict[key]
                other_item = other_dict[key]

                if datetime_fields.get(key, False) and current_item is not None:
                    # compare the values from the two dicts, with both shifted to UTC.
                    # any error here will be taken as the two instances are not equal
                    # see the except-branch
                    if current_item.astimezone(timezone.utc) != other_item.astimezone(timezone.utc):
                        return False
                elif current_item != other_item:
                    return False
            except:
                # in case of any error, return False.
                return False

        return True



"""Tests for benchmarks"""
import pytest

from test.conftest import book_collection_fixture, books, books_fixture, update_books_fixture, delete_books_fixture


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_add_many(benchmark, collection):
    """Benchmarks the add_many operation"""
    benchmark(collection.add_many, books)


@pytest.mark.parametrize("collection, book", books_fixture)
def test_benchmark_add_one(benchmark, collection, book):
    """Benchmarks the add_one() operation"""
    benchmark(collection.add_one, book)


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_get_all(benchmark, collection):
    """Benchmarks the get_all() operation"""
    collection.add_many(books)
    benchmark(collection.get_all)


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_get_all_partially(benchmark, collection):
    """Benchmarks the get_all_partially() operation"""
    collection.add_many(books)
    benchmark(collection.get_all_partially, fields=['title', 'author', 'in_stock'])


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_get_many(benchmark, collection):
    """Benchmarks the get_many() operation"""
    collection.add_many(books)
    ids = [book.title for book in books[:2]]
    benchmark(collection.get_many, ids=ids)


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_get_many_partially(benchmark, collection):
    """Benchmarks the get_many_partially() operation"""
    collection.add_many(books)
    ids = [book.title for book in books[:2]]
    benchmark(collection.get_many_partially, ids=ids, fields=['title', 'author', 'in_stock'])


@pytest.mark.parametrize("collection, book", books_fixture)
def test_benchmark_get_one_partially(benchmark, collection, book):
    """Benchmarks the get_one_partially() operation"""
    collection.add_many(books)
    benchmark(collection.get_one_partially, id=book.title, fields=['title', 'author', 'in_stock'])


@pytest.mark.parametrize("collection, book", books_fixture)
def test_benchmark_get_one(benchmark, collection, book):
    """Benchmarks the get_one() operation"""
    collection.add_many(books)
    benchmark(collection.get_one, id=book.title)


@pytest.mark.parametrize("collection, title, data", update_books_fixture)
def test_benchmark_update_one(benchmark, collection, title, data):
    """Benchmarks the update_one() operation"""
    collection.add_many(books)
    benchmark(collection.update_one, title, data=data)


@pytest.mark.parametrize("collection, title", delete_books_fixture)
def test_benchmark_delete(benchmark, collection, title):
    """Benchmarks the delete_many() for one id only operation"""
    collection.add_many(books)
    benchmark(collection.delete_many, ids=[title])


@pytest.mark.parametrize("collection", book_collection_fixture)
def test_benchmark_bulk_delete(benchmark, collection):
    """Benchmarks the bulk delete_many() operation"""
    collection.add_many(books)
    benchmark(collection.delete_many, ids=[book.title for book in books])



"""Tests for the redis orm"""
from datetime import datetime, timezone, timedelta

import pytest

from orredis import Model
from test.conftest import Book, redis_store_fixture, books, authors, Author


def test_model_eq():
    """the custom eq should be able to handle datetime objects"""

    class Data(Model):
        size: int
        name: str

    class Task(Data):
        time: datetime

    ts = datetime(year=2022, month=7, day=9, minute=8, second=7, tzinfo=timezone(timedelta(hours=5)))
    other_ts = datetime(year=2022, month=7, day=9, minute=8, second=7, tzinfo=timezone(timedelta(hours=2)))

    assert Data(size=9, name="foo") == Data(size=9, name="foo")
    assert Data(size=9, name="foo") != Data(size=9, name="foor")
    assert Data(size=9, name="foo") != Data(size=59, name="foo")

    assert Data(size=9, name="foo") == dict(size=9, name="foo")
    assert Data(size=9, name="foo") != dict(size=9, name="foor")
    assert Data(size=9, name="foo") != dict(size=59, name="foo")

    assert Task(time=ts, size=9, name="foo") == Task(time=ts.astimezone(timezone.utc), size=9, name="foo")
    assert Task(time=ts, size=9, name="foo") == Task(time=ts, size=9, name="foo")
    assert Task(time=ts, size=9, name="foo") != Task(time=ts.astimezone(timezone.utc), size=9, name="foor")
    assert Task(time=ts, size=9, name="foo") != Task(time=ts.astimezone(timezone.utc), size=59, name="foo")
    assert Task(time=ts, size=9, name="foo") != Task(time=other_ts, size=9, name="foo")

    assert Task(time=ts, size=9, name="foo") == dict(time=ts.astimezone(timezone.utc), size=9, name="foo")
    assert Task(time=ts, size=9, name="foo") == dict(time=ts, size=9, name="foo")
    assert Task(time=ts, size=9, name="foo") != dict(time=ts.astimezone(timezone.utc), size=9, name="foor")
    assert Task(time=ts, size=9, name="foo") != dict(time=ts.astimezone(timezone.utc), size=59, name="foo")
    assert Task(time=ts, size=9, name="foo") != dict(time=other_ts, size=9, name="foo")


def test_model_with_changes():
    """with_changes creates a new instance with the right modifications"""
    book = books[0]
    changes = {"title": "Yooo another book", "in_stock": not book.in_stock}
    new_book = book.with_changes(changes=changes)
    assert book != new_book
    for k, v in changes.items():
        value = getattr(new_book, k)
        assert value == v


def test_create_collection_without_primary_key(redis_store):
    """Throws error when a collection is created without a primary_kry_field"""

    class ModelWithoutPrimaryKey(Model):
        title: str

    with pytest.raises(TypeError, match=r"primary_key_field"):
        redis_store.create_collection(ModelWithoutPrimaryKey)

    with pytest.raises(TypeError,
                       match=r"argument 'primary_key_field': 'int' object cannot be converted to 'PyString'"):
        redis_store.create_collection(ModelWithoutPrimaryKey, primary_key_field=3)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_store_clear(store):
    """Clears all the keys in the redis store"""
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)

    book_collection.add_many(books)
    books_in_store_before_clear = book_collection.get_all()
    authors_in_store_before_clear = author_collection.get_all()

    store.clear()

    books_in_store_after_clear = book_collection.get_all()
    authors_in_store_after_clear = author_collection.get_all()

    assert books_in_store_before_clear != []
    assert authors_in_store_before_clear != []
    assert books_in_store_after_clear == []
    assert authors_in_store_after_clear == []


@pytest.mark.parametrize("store", redis_store_fixture)
def test_add_many(store):
    """Adds many model instances into the redis data store"""
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    books_in_store = book_collection.get_all()
    print(f"books: {[f'{bk}' for bk in books]}\n\nbooks_in_store: {[f'{bk}' for bk in books_in_store]}")
    assert sorted(books, key=lambda x: x.title) == sorted(books_in_store, key=lambda x: x.title)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_nested_add_many(store):
    """add_many also upserts any nested records in redis"""
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)

    authors_in_store_before_insert = author_collection.get_all()

    book_collection.add_many(books)

    authors_in_store_after_insert = sorted(author_collection.get_all(), key=lambda x: x.name)

    assert authors_in_store_before_insert == []
    assert authors_in_store_after_insert == sorted(authors.values(), key=lambda x: x.name)


@pytest.mark.parametrize("store", redis_store_fixture)
def test_add_one(store):
    """
    add_one inserts a single record into redis
    """
    book_collection = store.get_collection(Book)

    book = book_collection.get_one(id=books[0].title)
    assert book is None

    book_collection.add_one(books[0])

    book = book_collection.get_one(id=books[0].title)
    assert books[0] == book


@pytest.mark.parametrize("store", redis_store_fixture)
def test_nested_add_one(store):
    """
    add_one also any nested model into redis
    """
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)

    key = books[0].author.name
    author = author_collection.get_one(id=key)
    assert author is None

    book_collection.add_one(books[0])

    author = author_collection.get_one(id=key)
    assert books[0].author == author


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_all(store):
    """get_all() returns all the book models"""
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    response = book_collection.get_all()
    sorted_books = sorted(books, key=lambda x: x.title)
    sorted_response = sorted(response, key=lambda x: x.title)
    assert sorted_books == sorted_response


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_all_partially(store):
    """
    get_all_partially() returns a list of dictionaries of all books models with only those columns
    """
    book_collection = store.get_collection(Book)

    book_collection.add_many(books)
    books_dict = {book.title: book for book in books}
    columns = ['title', 'author', 'in_stock']
    response = book_collection.get_all_partially(fields=['title', 'author', 'in_stock'])
    response_dict = {book['title']: book for book in response}

    for title, book in books_dict.items():
        book_in_response = response_dict[title]
        assert isinstance(book_in_response, dict)
        assert sorted(book_in_response.keys()) == sorted(columns)

        for column in columns:
            if column == 'author':
                assert book_in_response[column] == getattr(book, column)
            else:
                v = getattr(book, column)
                assert f"{book_in_response[column]}" == f"{v}"


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_many_partially(store):
    """
    get_many_partially() returns a list of dictionaries of book models of the selected ids
    with only those fields
    """
    book_collection = store.get_collection(Book)

    book_collection.add_many(books)
    ids = [book.title for book in books[:2]]
    books_dict = {book.title: book for book in books[:2]}
    fields = ['title', 'author', 'in_stock']
    response = book_collection.get_many_partially(ids=ids, fields=['title', 'author', 'in_stock'])
    response_dict = {book['title']: book for book in response}

    assert len(response) == len(ids)

    for title, book in books_dict.items():
        book_in_response = response_dict[title]
        assert isinstance(book_in_response, dict)
        assert sorted(book_in_response.keys()) == sorted(fields)

        for column in fields:
            if column == 'author':
                assert book_in_response[column] == getattr(book, column)
            else:
                v = getattr(book, column)
                assert f"{book_in_response[column]}" == f"{v}"


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_many(store):
    """
    get_many() returns only those elements with the given ids
    """
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    ids = [book.title for book in books[:2]]
    response = book_collection.get_many(ids=ids)
    assert response == books[:2]


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_one_non_existent_id(store):
    """
    get_one() for a non-existent id returns None
    """
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    response = book_collection.get_one(id="Some strange book")
    assert response is None


# FIXME: add test for non-existent columns for both single and multiple record retrieval

@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_one_partially_non_existent_id_with_existent_columns(store):
    """
    get_one_partially() for a non-existent id even when columns are okay returns None
    """
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    response = book_collection.get_one_partially(id="Some strange book", fields=["author", "title"])
    assert response is None


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_one(store):
    """
    get_one() returns only the elements with the given id
    """
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    for book in books:
        response = book_collection.get_one(id=book.title)
        assert response == book


@pytest.mark.parametrize("store", redis_store_fixture)
def test_get_one_partially(store):
    """
    get_one_partially() returns only the fields given for the elements with the given id
    """
    book_collection = store.get_collection(Book)
    book_collection.add_many(books)
    fields = ['title', "author", 'in_stock']

    for book in books:
        response = book_collection.get_one_partially(id=book.title, fields=fields)
        expected = {key: getattr(book, key) for key in fields}
        assert expected == response


@pytest.mark.parametrize("store", redis_store_fixture)
def test_update_one(store):
    """
    update_one() for a given primary key updates it in redis
    """
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)
    book_collection.add_many(books)
    title = books[0].title
    new_in_stock = not books[0].in_stock
    new_author = Author(name='John Doe', active_years=(2000, 2009))
    new_author_key = new_author.name

    old_book = book_collection.get_one(id=title)
    assert old_book == books[0]
    assert old_book.author != new_author

    book_collection.update_one(id=title, data={"author": new_author, "in_stock": new_in_stock})

    book = book_collection.get_one(id=title)
    author = author_collection.get_one(id=new_author_key)
    assert book.author == new_author
    assert author == new_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.parametrize("store", redis_store_fixture)
def test_nested_update_one(store):
    """
    Updating a nested model, without changing its primary key, also updates it its collection in redis
    """
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)
    book_collection.add_many(books)

    new_in_stock = not books[0].in_stock
    updated_author = Author(**books[0].author.dict())
    updated_author.active_years = (2020, 2045)
    book_key = books[0].title
    author_key = updated_author.name

    old_author = author_collection.get_one(id=author_key)
    old_book = book_collection.get_one(id=book_key)
    assert old_book == books[0]
    assert old_author == books[0].author
    assert old_author != updated_author

    book_collection.update_one(id=books[0].title, data={"author": updated_author, "in_stock": new_in_stock})

    book = book_collection.get_one(id=book_key)
    author = author_collection.get_one(id=author_key)

    assert book.author == updated_author
    assert author == updated_author
    assert book.title == old_book.title
    assert book.in_stock == new_in_stock
    assert book.published_on == old_book.published_on


@pytest.mark.parametrize("store", redis_store_fixture)
def test_delete_many(store):
    """
    delete_many() removes the items of the given ids from redis,
    but leaves the nested models intact
    """
    book_collection = store.get_collection(Book)
    author_collection = store.get_collection(Author)
    book_collection.add_many(books)
    books_to_delete = books[:2]
    books_to_be_left_in_db = books[2:]

    ids_to_delete = [book.title for book in books_to_delete]
    ids_to_leave_intact = [book.title for book in books_to_be_left_in_db]

    book_collection.delete_many(ids=ids_to_delete)
    deleted_books_select_response = book_collection.get_many(ids=ids_to_delete)

    books_left = book_collection.get_many(ids=ids_to_leave_intact)
    authors_left = sorted(author_collection.get_all(), key=lambda x: x.name)

    assert deleted_books_select_response == []
    assert books_left == books_to_be_left_in_db
    assert authors_left == sorted(authors.values(), key=lambda x: x.name)



from typing import Optional, Type, List, Dict, Any

from .abstract import Model


class Collection:
    """
    The Collection represents a group of similar records within redis
    e.g. records of model Car will be in the Car collection.
    This has all the methods for manipulating data as well as retrieving it
    from redis
    """

    def add_one(self, item: Model, ttl: Optional[int]) -> None:
        """
        Adds a single Model instance to the collection. The model instance should
        be of the same type as the model used to get this collection.

        :param item: the model object to add
        :param ttl: the optional time-to-live for this item in redis; default: None (i.e. never expires).
                    if however, the default_ttl was set on the store, it will default to that
        """

    def add_many(self, items: List[Model], ttl: Optional[int]) -> None:
        """
        Adds a list of Model instances into this collection. The model instances should
        be of the same type as the model used to get this collection.

        :param items: the list of model objects to add
        :param ttl: the optional time-to-live for all these items in redis; default: None (i.e. never expires)
                    if however, the default_ttl was set on the store, it will default to that
        """

    def update_one(self, id: str, data: Dict[str, Any], ttl: Optional[int]) -> None:
        """
        Updates the model instance in redis that has the given id

        :param id: the id of the record to update
        :param data: the new changes to add to the record
        :param ttl: the optional time-to-live for all these items in redis; default: None (i.e. never expires)
                    if however, the default_ttl was set on the store, it will default to that
        """

    def get_one(self, id: str) -> Model:
        """
        Retrieves one record of the given id or None if it does not exist

        :param id: the id of the model record to return
        :return: the model object for the given id in this collection
        """

    def get_many(self, ids: List[str]) -> List[Model]:
        """
        Retrieves a list of records in this collection corresponding to the ids passed

        :param ids: the list of ids whose records are to be returned
        :return: the list of model objects that correspond to the list of ids. Non-existent ids
               are simply skipped
        """

    def get_all(self) -> List[Model]:
        """
        Retrieves a list of all records in this collection at ago

        :return: the list of model objects in this collection
        """

    def get_one_partially(self, id: str, fields: List[str]) -> Dict[str, Any]:
        """
        Retrieves a dictionary containing the provided fields from the record of the given id

        :param id: the id of the record whose data is to be retrieved
        :param fields: the list of fields to be returned in the data
        :return: the dict with the given fields as keys and the values got from the record of the given id
        """

    def get_many_partially(self, ids: List[str], fields: List[str]) -> List[Dict[str, Any]]:
        """
        Retrieves a list of dictionaries for records of the given ids,
        only returning the specified fields for each record

        :param ids: the list of ids of the records to be queried
        :param fields: the fields to be returned in each item
        :return: the list of dicts, each with the given fields as keys and the values for each record returned.
               non-existent ids are ignored
        """

    def get_all_partially(self, fields: List[str]) -> List[Dict[str, Any]]:
        """
        Retrieves a list of dictionaries for all records in the store,
        only returning the specified fields for each record

        :param fields: the fields to be returned in each item
        :return: the list of dicts, each with the given fields as keys and the values for each record returned
        """

    def delete_many(self, ids: List[str]) -> None:
        """
        Removes all records belonging to the given ids

        :param ids: the ids of the records to be removed
        """


class Store:
    """
    The Store containing all collections that are stored in redis.

    :param url: the redis url e.g. redis://localhost:6379/0
    :param pool_size: the maximum number of connections in the connection pool to redis; default: 5
    :param default_ttl: the default time-to-live for each record in milliseconds; default: None i.e. no expiry
    :param timeout: the time in milliseconds beyond which a timeout error is raised on failure to
                    get a connection to redis from the connection pool; default is 30000 (30 seconds)
    :param max_lifetime: the maximum lifetime in milliseconds connections in the pool; default is 1800000 (30 minutes)
    """

    def __init__(self,
                 url: str,
                 pool_size: int,
                 default_ttl: Optional[int],
                 timeout: Optional[int],
                 max_lifetime: Optional[int]) -> None: ...

    def clear(self, asynchronous: bool = False) -> None:
        """
        Removes all records in the redis store

        :param asynchronous: whether the FLUSHALL should be done asynchronously or synchronously. default: False
        """

    def create_collection(self,
                          model: Type[Model],
                          primary_key_field: str) -> None:
        """
        Creates a new Collection within the store for the given model supplied

        :param model: the Model schema to be used for this collection
        :param primary_key_field: the field that contains the unique primary key for each model instance e.g.
                                a book's primary key might be its ISBN
        """

    def get_collection(self, model: Type[Model]) -> Collection:
        """
        Retrieves a handle on the collection for a given model to manipulate the data within or
        query it

        :param model: the Model schema whose collection is to be retrieved
        :return: the collection instance to be used to manipulate data or query it using collection.add_one() etc.
        """



import socket
from datetime import date, datetime, timezone
from typing import Tuple, List

import pytest
import redislite
from pytest_lazyfixture import lazy_fixture

from orredis import Store, Model


class Author(Model):
    name: str
    active_years: Tuple[int, int]

class Book(Model):
    title: str
    author: Author
    rating: float
    published_on: date
    last_updated: datetime = datetime(year=2022, month=9, day=17, hour=1, minute=30, tzinfo=timezone.utc)
    tags: List[str] = []
    in_stock: bool = True

authors = {
    "charles": Author(name="Charles Dickens", active_years=(1220, 1280)),
    "jane": Author(name="Jane Austen", active_years=(1580, 1640)),
}

books = [
    Book(title="Oliver Twist", author=authors["charles"], published_on=date(year=1215, month=4, day=4),
         in_stock=False, rating=2, tags=["Classic"]),
    Book(title="Great Expectations", author=authors["charles"], published_on=date(year=1220, month=4, day=4), rating=5,
         tags=["Classic"]),
    Book(title="Jane Eyre", author=authors["charles"], published_on=date(year=1225, month=6, day=4), in_stock=False,
         rating=3.4, tags=["Classic", "Romance"]),
    Book(title="Wuthering Heights", author=authors["jane"], published_on=date(year=1600, month=4, day=4), rating=4.0,
         tags=["Classic", "Romance"]),
]

redis_store_fixture = [(lazy_fixture("redis_store"))]
book_collection_fixture = [(lazy_fixture("book_collection"))]
books_fixture = [(lazy_fixture("book_collection"), book) for book in books[-1:]]
update_books_fixture = [
    (lazy_fixture("book_collection"), book.title, {"author": authors["jane"], "in_stock": not book.in_stock})
    for book in books[-1:]
]
delete_books_fixture = [(lazy_fixture("book_collection"), book.title) for book in books[-1:]]


@pytest.fixture()
def unused_tcp_port():
    """Creates an unused TCP port"""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    port = f"{sock.getsockname()[1]}"
    sock.close()
    yield port


@pytest.fixture()
def redis_server(unused_tcp_port):
    """Sets up a fake redis server we can use for tests"""
    instance = redislite.Redis(serverconfig={"port": unused_tcp_port})
    yield unused_tcp_port
    instance.shutdown()


@pytest.fixture()
def redis_store(redis_server):
    """Sets up a redis store using the redis_server fixture and adds the book model to it"""
    store = Store(url=f"redis://localhost:{redis_server}/1")
    store.create_collection(Author, primary_key_field="name")
    store.create_collection(Book, primary_key_field="title")
    yield store
    store.clear()


@pytest.fixture()
def book_collection(redis_store):
    """Returns a collection for manipulating book records"""
    yield redis_store.get_collection(Book)


@pytest.fixture()
def author_collection(redis_store):
    """Returns a collection for manipulating author records"""
    yield redis_store.get_collection(Author)
